/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifndef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gir-lirc.h"
#include "gir-marshal.h"

#include <glib/gi18n.h>
#include <glib/gfileutils.h>

#include <errno.h>

#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>

enum {
	RECV_IR,
	DISCONNECTED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

typedef enum {
	STATE_NORMAL,
	STATE_REPLY_BEGIN,
	STATE_REPLY_COMMAND,
	STATE_REPLY_RESULT,
	STATE_REPLY_DATA,
	STATE_REPLY_NLINE
} State;

typedef struct {
	GIOChannel *iochannel;

	GList *commands;

	State state;
	char *command;
	gboolean result;
	int lines;
	GList *reply;
} Private;

#define GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), GIR_TYPE_LIRC, Private))

G_DEFINE_TYPE(GirLirc, gir_lirc, G_TYPE_OBJECT)



static gboolean
on_input (GIOChannel *iochannel, GIOCondition condition, gpointer data)
{
	GirLirc *lirc = GIR_LIRC (data);
	Private *priv = GET_PRIVATE (lirc);
	char *line;
	GIOStatus status;

	/* FIXME: error processing */

	/*
	 * IR signal:
	 *
	 * <code> <repeat count> <button name> <control name>
	 *
	 * Reply from LIRCD:
	 *
	 * BEGIN
	 * <command>
	 * [SUCCESS|ERROR]
	 * [DATA
	 * n
	 * n lines of data]
	 * END
	 */
	while (TRUE) {
		status = g_io_channel_read_line (iochannel, &line,
						 NULL, NULL, NULL);
		if (status != G_IO_STATUS_NORMAL)
			break;
		g_strchomp (line);
		switch (priv->state) {
		case STATE_NORMAL:
			if (strncmp (line, "BEGIN", 5) == 0) {
				priv->command = NULL;
				priv->reply = NULL;
				priv->state = STATE_REPLY_BEGIN;
			}
			else {
				char **strv;

				strv = g_strsplit (line, " ", 4);
				if (!strv)
					break;
				if (strv[0] && strv[1] && strv[2] && strv[3]) {
					int repeat;
					char *button, *control;

					repeat = strtoul(strv[1], NULL, 16);
					button = strv[2];
					control = strv[3];

					g_signal_emit (lirc,
						       signals[RECV_IR], 0,
						       repeat,
						       button,
						       control);
				}
				g_strfreev (strv);
			}
			break;
		case STATE_REPLY_BEGIN:
			priv->command = line;
			line = NULL;
			priv->state = STATE_REPLY_COMMAND;
			break;
		case STATE_REPLY_COMMAND:
			if (strncmp (line, "SUCCESS", 6) == 0) {
				priv->result = TRUE;
				priv->state = STATE_REPLY_RESULT;
			}
			else if (strncmp (line, "ERROR", 5) == 0) {
				priv->result = FALSE;
				priv->state = STATE_REPLY_RESULT;
			}
			else {
				g_warning ("Wrong repy from LIRCD");
				g_free (priv->command);
				priv->state = STATE_NORMAL;
			}
			break;
		case STATE_REPLY_RESULT:
			if (strncmp (line, "DATA", 4) == 0) {
				priv->state = STATE_REPLY_DATA;
			}
			else {
				g_warning ("Wrong repy from LIRCD");
				g_free (priv->command);
				priv->state = STATE_NORMAL;
			}
			break;
		case STATE_REPLY_DATA:
			priv->lines = atoi (line);
			priv->state = STATE_REPLY_NLINE;
			break;
		case STATE_REPLY_NLINE:
			if (priv->lines) {
				priv->reply = g_list_append (priv->reply,
							     line);
				priv->lines--;
				line = NULL;
			}
			else if (strncmp (line, "END", 3) == 0) {
				g_main_loop_quit (priv->commands->data);
				priv->state = STATE_NORMAL;
			}
			else {
				GList *l; 

				g_warning ("Wrong reply from LIRCD");
				g_free (priv->command);
				priv->state = STATE_NORMAL;
				for (l = priv->reply; l; l = l->next) {
					g_free (l->data);
				}
				g_list_free (priv->reply);
				priv->state = STATE_NORMAL;
			}					
		}

		if (line)
			g_free (line);
	}

	return TRUE;
}

static gboolean
on_hangup (GIOChannel *iochannel, GIOCondition condition, gpointer data)
{
	GirLirc *lirc = GIR_LIRC (data);
	gir_lirc_disconnect (lirc);
	g_signal_emit (lirc, signals[DISCONNECTED], 0);
	return FALSE;
}




static void
gir_lirc_init (GirLirc *lirc)
{
	Private *priv = GET_PRIVATE (lirc);
	priv->state = STATE_NORMAL;
}

static void
gir_lirc_class_init (GirLircClass *klass)
{
	g_type_class_add_private (klass, sizeof (Private));

	signals[RECV_IR] =
		g_signal_new ("recv_ir",
			      G_OBJECT_CLASS_TYPE (klass),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (GirLircClass, recv_ir),
			      NULL, NULL,
			      _gir_marshal_VOID__INT_STRING_STRING,
			      G_TYPE_NONE,
			      3, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);
	signals[DISCONNECTED] =
		g_signal_new ("disconnected",
			      G_OBJECT_CLASS_TYPE (klass),
			      G_SIGNAL_RUN_FIRST,
			      G_STRUCT_OFFSET (GirLircClass, recv_ir),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__POINTER,
			      G_TYPE_NONE,
			      1, G_TYPE_POINTER);
}

GirLirc*
gir_lirc_new (void)
{
	return g_object_new (GIR_TYPE_LIRC, NULL);
}



gboolean
gir_lirc_is_connected (GirLirc *lirc)
{
	Private *priv = GET_PRIVATE (lirc);

	return (priv->iochannel != NULL);
}

gboolean
gir_lirc_connect (GirLirc *lirc, GError **error)
{
	Private *priv = GET_PRIVATE (lirc);
	struct sockaddr_un addr;
	int sock;

	g_return_val_if_fail (priv->iochannel == NULL, FALSE);

	sock = socket (AF_UNIX, SOCK_STREAM, 0);
	if (sock < 0) {
		if (error) {
			g_set_error (error,
				     G_FILE_ERROR,
				     g_file_error_from_errno (errno),
				     _("Connecting to LIRC daemon failed: %s"),
				     g_strerror (errno));
		}
		return FALSE;
	}

	addr.sun_family=AF_UNIX;
	strcpy (addr.sun_path, "/dev/lircd");
	if (connect (sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		if (error) {
			g_set_error (error,
				     G_FILE_ERROR,
				     g_file_error_from_errno (errno),
				     _("Connecting to LIRC daemon failed: %s"),
				     g_strerror (errno));
		}
		close (sock);
		return FALSE;
	}

	priv->iochannel = g_io_channel_unix_new (sock);
	g_io_channel_set_flags (priv->iochannel, G_IO_FLAG_NONBLOCK, NULL);

	g_io_add_watch (priv->iochannel, G_IO_IN, on_input, lirc);
	g_io_add_watch (priv->iochannel, G_IO_HUP, on_hangup, lirc);
	return TRUE;
}

void
gir_lirc_disconnect (GirLirc *lirc)
{
	Private *priv = GET_PRIVATE (lirc);

	if (priv->iochannel) {
		g_io_channel_shutdown (priv->iochannel, TRUE, NULL);
		g_io_channel_unref (priv->iochannel);
		priv->iochannel = NULL;
	}
}

char *
gir_lirc_get_lirc_version (GirLirc *lirc)
{
	Private *priv = GET_PRIVATE (lirc);
	char *vstring;
	GMainLoop *main_loop;
	int nwritten;

	g_io_channel_write (priv->iochannel, "VERSION\n", 8, &nwritten);

	main_loop = g_main_loop_new (g_main_context_default (), FALSE);
	priv->commands = g_list_append (priv->commands, main_loop);
	g_main_loop_run (main_loop);
	g_main_loop_unref (main_loop);
	priv->commands = g_list_delete_link (priv->commands, priv->commands);

	vstring = priv->reply->data;
	g_list_free (priv->reply);
	return vstring;
}

GList *
gir_lirc_get_button_list (GirLirc *lirc, const char *control)
{
	Private *priv = GET_PRIVATE (lirc);
	GMainLoop *main_loop;
	int nwritten;
	char *request;
	GList *l;

	request = g_strdup_printf ("LIST %s\n", control);
	g_io_channel_write (priv->iochannel, request, strlen (request),
			    &nwritten);
	g_free (request);

	main_loop = g_main_loop_new (g_main_context_default (), FALSE);
	priv->commands = g_list_append (priv->commands, main_loop);
	g_main_loop_run (main_loop);
	g_main_loop_unref (main_loop);
	priv->commands = g_list_delete_link (priv->commands, priv->commands);

	for (l = priv->reply; l; l = l->next) {
		char *p = strchr (l->data, ' ');
		if (p)
			g_memmove (l->data, p + 1, strlen (p + 1) + 1);
	}

	return priv->reply;
}

GList *
gir_lirc_get_control_list (GirLirc *lirc)
{
	Private *priv = GET_PRIVATE (lirc);
	GMainLoop *main_loop;
	int nwritten;

	g_io_channel_write (priv->iochannel, "LIST\n", 5, &nwritten);

	main_loop = g_main_loop_new (g_main_context_default (), FALSE);
	priv->commands = g_list_append (priv->commands, main_loop);
	g_main_loop_run (main_loop);
	g_main_loop_unref (main_loop);
	priv->commands = g_list_delete_link (priv->commands, priv->commands);

	return priv->reply;
}

