/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
#include "gir-mappings.h"

#include <glib/gi18n.h>
#include <gconf/gconf-client.h>
#include <gtk/gtktreestore.h>
#include <libxml/parser.h>

#include <string.h>

struct _GirMappings
{
	GtkTreeStore *tree_store;
	GHashTable *action_hash;   /* action for each control+button */

	/* gconf sync */
	guint gconf_cnxn_id;
};

#define GCONF_KEY_MAPPINGS "/apps/gir/mappings"




typedef struct {
	char *control;
	char *button;
} ActionKey;



typedef void (*ActionFn) (void);

static void do_action_key (void);

static const struct {
	GirActionType	type;
	char	       *repr;	/* string representation */
	ActionFn	func;
} actions[] = {
	{ GIR_ACTION_NONE,	N_("Disabled"),	NULL		},
	{ GIR_ACTION_KEY,	N_("Key %s"),	do_action_key	},
};



static GirAction *
action_copy (const GirAction *src)
{
	GirAction *copy;

	copy = g_new (GirAction, 1);

	copy->type = src->type;
	switch (copy->type) {
	case GIR_ACTION_NONE:
		break;
	case GIR_ACTION_KEY:
		copy->data1 = g_strdup (src->data1);
		break;
	}
	return copy;
}

static void
action_free (GirAction *a)
{
	switch (a->type) {
	case GIR_ACTION_NONE:
		break;
	case GIR_ACTION_KEY:
		g_free (a->data1);
		break;
	}
}



static guint
hash_func (gconstpointer key)
{
	ActionKey *k = (ActionKey *) key;
	const char *p;
	guint h;

	p = k->control;
	h = 0;

	for (p = k->control; *p; p++) {
		h = (h << 5) - h + *p;
	}
	for (p = k->button; *p; p++) {
		h = (h << 5) - h + *p;
	}
	return h;
}

static gboolean
equal_func (gconstpointer a, gconstpointer b)
{
	ActionKey *ka = (ActionKey *) a;
	ActionKey *kb = (ActionKey *) b;

	if (strcmp (ka->control, kb->control) != 0)
		return FALSE;
	return (strcmp (ka->button, kb->button) == 0);
}

static void
destroy_key (gpointer key)
{
	ActionKey *ak = (ActionKey *)key;
	g_free (ak->control);
	g_free (ak->button);
	g_free (ak);
}

static void
destroy_value (gpointer value)
{
	action_free (value);
}

#if 0
static void
insert_xmlnode (gpointer key, gpointer value, gpointer data)
{
	xmlNodePtr root = (xmlNodePtr)data;
	xmlNodePtr ctrl, btn, c;
	xmlNodePtr new_btn, new_action;
	ActionKey *ak = (ActionKey*) key;
	ActionData *ad = (ActionData*) value;

	new_btn = xmlNewNode (NULL, "button");

	ctrl = root->children;
	for (; ctrl; ctrl = ctrl->next) {
		if (strcmp (ctrl->name, ak->control) == 0) {
			c = ctrl;
			goto control_found;
		} else if (strcmp (ctrl->name, ak->control) < 0) {
			c = xmlNewNode (NULL, "control");
			xmlAddPrevSibling (ctrl, c);
			goto control_found;
		}
	}
	c = xmlNewChild (root, NULL, "control", NULL);

 control_found:

	btn = c->children;
	for (; btn; btn = btn->next) {
		if (strcmp (btn->name, ak->button) == 0) {
			/* remove existing actions */
			if (btn->children)
				xmlUnlinkNode (btn->children);
			/* let's update it. */
			xmlFreeNode (new_btn);
			new_btn = btn;
			goto button_found;
		}
		else if (strcmp (btn->name, ak->button) < 0) {
			xmlAddPrevSibling (btn, new_btn);
			goto button_found;
		}
	}
	xmlAddChild (c, new_btn);
 button_found:

	switch (ad->type) {
	case GIR_ACTION_NONE:	/* no action */
		break;
	case GIR_ACTION_KEY:
		new_action = xmlNewChild (new_btn, NULL, "key", ad->key_value);
		break;
	}
}
#endif



static void
do_action_key ()
{
}



static gboolean
load_button_xmlnode (GirMappings *mappings, const char *ctrl,
		     GtkTreeIter *parent, GtkTreeIter *before, GtkTreeIter *result,
		     xmlNodePtr node)
{
	GtkTreeStore *tree_store = GTK_TREE_STORE (mappings->tree_store);
	char *name;
	xmlNodePtr actionnode;
	ActionKey *ak;
	GirAction *action;
	char *action_string;

	if (strcmp (node->name, "button") != 0) {
		g_warning ("not <button> tag");
		return FALSE;
	}
	name = xmlGetProp (node, "name");
	if (!name) {
		g_warning ("no \"name\" attribute in <button> tag");
		return FALSE;
	}

	gtk_tree_store_insert_after (tree_store, result, parent, before);
	gtk_tree_store_set (tree_store, result, 0, name, -1);
	
	ak = g_new (ActionKey, 1);
	ak->control = g_strdup (ctrl);
	ak->button = g_strdup (name);
	action = g_new (GirAction, 1);
	
	actionnode = node->children;
	if (!actionnode) {
		action->type = GIR_ACTION_NONE;
	}
	else if (strcmp (actionnode->name, "key") == 0 &&
		 actionnode->children &&
		 actionnode->children->type == XML_TEXT_NODE) {
		action->type = GIR_ACTION_KEY;
		action->data1 = g_strdup (XML_GET_CONTENT (actionnode->children));
	}
	else {
		action->type = GIR_ACTION_NONE;
	}

	g_hash_table_insert (mappings->action_hash, ak, action);

	action_string = g_strdup_printf (actions[action->type].repr, action->data1);
	gtk_tree_store_set (tree_store, result, 1, action_string, -1);
	g_free (action_string);

	xmlFree (name);
	return TRUE;
}

static gboolean
load_control_xmlnode (GirMappings *mappings, GtkTreeIter *before, GtkTreeIter *result,
		      xmlNodePtr node)
{
	GtkTreeStore *tree_store = GTK_TREE_STORE (mappings->tree_store);
	char *name;
	xmlNodePtr btn;
	GtkTreeIter biter, *biterp;

	if (strcmp (node->name, "control") != 0)
		return FALSE;
	name = xmlGetProp (node, "name");
	if (!name) {
		g_warning ("no \"name\" attribute in <control> tag");
		return FALSE;
	}

	gtk_tree_store_insert_after (tree_store, result, NULL, before);
	gtk_tree_store_set (tree_store, result, 0, name, -1);

	biterp = NULL;
	for (btn = node->children; btn; btn = btn->next) {
		load_button_xmlnode (mappings, name,
				     result, biterp, &biter, btn);
		biterp = &biter;
	}

	xmlFree (name);
	return FALSE;
}

static gboolean
load_xmldoc (GirMappings *mappings, xmlDocPtr doc)
{
	GtkTreeModel *tree_store = GTK_TREE_MODEL (mappings->tree_store);
	xmlNodePtr root, ctrl;
	GtkTreeIter iter, *iterp;
	gboolean valid;

	/* remove all */
	valid = gtk_tree_model_get_iter_first (tree_store, &iter);
	while (valid)
		valid = gtk_tree_store_remove (GTK_TREE_STORE (tree_store),
					       &iter);
	/* FIXME: remove hash data */
	
	root = xmlDocGetRootElement (doc);
	if (!root)
		return FALSE;
	if (strcmp (root->name, "mappings") != 0) {
		g_warning ("root is not <mappings>");
		return FALSE;
	}

	iterp = NULL;
	for (ctrl = root->children; ctrl; ctrl = ctrl->next) {
		load_control_xmlnode (mappings, iterp, &iter, ctrl);
		iterp = &iter;
	}

	return TRUE;
}



static void
setup_changed (GConfClient *client, guint cnxn_id,
	       GConfEntry *entry, gpointer user_data)
{
	const char *xml;
	
	xml = gconf_value_get_string (gconf_entry_get_value (entry));
	gir_mappings_load_xml ((GirMappings*)user_data, xml);
}



GirMappings*
gir_mappings_new ()
{
	GirMappings *mappings;
	char *xml;

	mappings = g_new0 (GirMappings, 1);

	mappings->action_hash = g_hash_table_new_full (hash_func, equal_func,
						       destroy_key,
						       destroy_value);
	mappings->tree_store =
		gtk_tree_store_new (2, G_TYPE_STRING, G_TYPE_STRING);

	mappings->gconf_cnxn_id =
		gconf_client_notify_add (gconf_client_get_default (),
					 GCONF_KEY_MAPPINGS,
					 setup_changed,
					 mappings,
					 NULL,
					 NULL);

	xml = gconf_client_get_string (gconf_client_get_default (),
				       GCONF_KEY_MAPPINGS,
				       NULL);
	gir_mappings_load_xml (mappings, xml);
	g_free (xml);
	
	return mappings;
}

void
gir_mappings_free (GirMappings *mappings)
{
	gconf_client_notify_remove (gconf_client_get_default (),
				    mappings->gconf_cnxn_id);
	g_free (GCONF_KEY_MAPPINGS);

	g_hash_table_destroy (mappings->action_hash);
	g_object_unref (G_OBJECT (mappings->tree_store));
	g_free (mappings);
}

gboolean
gir_mappings_load_xml (GirMappings *mappings, const char *xml)
{
	gboolean ret;
	xmlDocPtr doc;

	doc = xmlParseDoc ((xmlChar*)xml);
	ret = load_xmldoc (mappings, doc);
	xmlFreeDoc (doc);
	return ret;
}



char *
gir_mappings_dump_xml (GirMappings *mappings)
{
	GtkTreeModel *tree_model = GTK_TREE_MODEL (mappings->tree_store);
	xmlDocPtr doc;
	xmlNodePtr root, cnode, bnode;
	gboolean cvalid, bvalid;
	GtkTreeIter citer, biter;
	char *cname, *bname;

	xmlChar *xml_buffer;
	int xml_buffer_size;
	char *return_buffer;


	doc = xmlNewDoc ("1.0");

	root = xmlNewDocNode (doc, NULL, "mappings", NULL);
	xmlDocSetRootElement (doc, root);

	cvalid = gtk_tree_model_get_iter_first (tree_model, &citer);
	while (cvalid) {
		cnode = xmlNewChild (root, NULL, "control", NULL);
		gtk_tree_model_get (tree_model, &citer, 0, &cname, -1);
		xmlSetProp (cnode, "name", cname);

		bvalid = gtk_tree_model_iter_children (tree_model, &biter,
						       &citer);
		while (bvalid) {
			bnode = xmlNewChild (cnode, NULL, "button", NULL);
			gtk_tree_model_get (tree_model, &biter, 0, &bname, -1);
			xmlSetProp (bnode, "name", bname);
			
			bvalid = gtk_tree_model_iter_next (tree_model, &biter);
		}

		g_free (cname);
		cvalid = gtk_tree_model_iter_next (tree_model, &citer);
	}

	xmlDocDumpMemory (doc, &xml_buffer, &xml_buffer_size);
	xmlFreeDoc (doc);

	return_buffer = g_malloc (xml_buffer_size + 1);
	memcpy (return_buffer, xml_buffer, xml_buffer_size);
	return_buffer [xml_buffer_size] = '\0';
	xmlFree (xml_buffer);

	return return_buffer;
}

const GirAction*
gir_mappings_get_action  (GirMappings *mappings,
			  const char *control,
			  const char *button)
{
	ActionKey ak;
	GirAction *ad;

	ak.control = (char*)control;
	ak.button = (char*)button;

	ad = g_hash_table_lookup (mappings->action_hash, &ak);

	return ad;
}

static void
insert_control_if_none (GtkTreeStore *tree_store, const char *control,
			GtkTreeIter *iter)
{
	GtkTreeModel *tree_model = GTK_TREE_MODEL (tree_store);
	char *name;
	gboolean valid;
	int compare;

	valid = gtk_tree_model_get_iter_first (tree_model, iter);
	while (valid) {
		gtk_tree_model_get (tree_model, iter, 0, &name, -1);
		compare = strcmp (control, name);
		g_free (name);
		if (compare == 0)
			return;
		else if (compare < 0)
			break;

		valid = gtk_tree_model_iter_next (tree_model, iter);
	}

	if (valid)
		gtk_tree_store_insert_before (tree_store, iter, NULL, iter);
	else
		gtk_tree_store_insert_before (tree_store, iter, NULL, NULL);
	gtk_tree_store_set (tree_store, iter, 0, control, -1);
}

static void
insert_button_if_none (GtkTreeStore *tree_store, GtkTreeIter *parent,
		       const char *button, GtkTreeIter *iter)
{
	GtkTreeModel *tree_model = GTK_TREE_MODEL (tree_store);
	char *name;
	gboolean valid;
	int compare;

	valid = gtk_tree_model_iter_children (tree_model, iter, parent);
	while (valid) {
		gtk_tree_model_get (tree_model, iter, 0, &name, -1);
		compare = strcmp (button, name);
		g_free (name);
		if (compare == 0)
			return;
		else if (compare < 0)
			break;

		valid = gtk_tree_model_iter_next (tree_model, iter);
	}

	if (valid)
		gtk_tree_store_insert_before (tree_store, iter, parent, iter);
	else
		gtk_tree_store_insert_before (tree_store, iter, parent, NULL);
	gtk_tree_store_set (tree_store, iter, 0, button, -1);
}

void
gir_mappings_set_action  (GirMappings *mappings,
			  const char *control,
			  const char *button,
			  const GirAction *action)
{
	GtkTreeStore *tree_store = GTK_TREE_STORE (mappings->tree_store);
	GtkTreeIter citer, biter;
	ActionKey *key;
	GirAction *a;
	char *action_string;

	key = g_new (ActionKey, 1);
	key->control = g_strdup (control);
	key->button = g_strdup (button);
	a = action_copy (action);

	g_hash_table_insert (mappings->action_hash, key, a);
	insert_control_if_none (mappings->tree_store, control, &citer);
	insert_button_if_none (mappings->tree_store, &citer, button, &biter);

	action_string = g_strdup_printf (actions[action->type].repr, action->data1);
	gtk_tree_store_set (tree_store, &biter, 1, action_string, -1);
	g_free (action_string);
}

void
gir_mappings_set_action_default  (GirMappings *mappings,
				  const char *control,
				  const char *button)
{
	GirAction *a;
	a = g_new (GirAction, 1);
	a->type = GIR_ACTION_NONE;

	gir_mappings_set_action (mappings, control, button, a);
}

void
gir_mappings_do_action (GirMappings *mappings,
			const char *control,
			const char *button,
			int repeat)
{
	GirAction *a;
	ActionKey key;

	key.control = (char*)control;
	key.button = (char*)button;

	a = g_hash_table_lookup (mappings->action_hash, &key);
	if (!a) {
		g_warning ("No action registered for control %s button %s",
			   control, button);
		return;
	}

	/* do something with it */
}

GtkTreeModel*
gir_mappings_get_tree_model (GirMappings *mappings)
{
	return GTK_TREE_MODEL (mappings->tree_store);
}


void
gir_mappings_sync (GirMappings *mappings)
{
	char *xml;

	xml = gir_mappings_dump_xml (mappings);
	gconf_client_set_string (gconf_client_get_default (),
				 GCONF_KEY_MAPPINGS,
				 xml,
				 NULL);
	g_free (xml);
}
