/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n.h>
#include "gir-preferences.h"

#include <gtk/gtk.h>
#include <glade/glade.h>

enum 
{
  SIGNAL_COLUMN = 0,
  ACTION_COLUMN
};

typedef struct {
	GladeXML *xml;

	GtkTreeModel *tree_model;

	GtkTreeView *tree_view;
	GtkCheckButton *test_check_button;
	GtkLabel *test_control_label;
	GtkLabel *test_button_label;
	GtkLabel *test_repeat_label;

	GirMappings *mappings;

	const GList* (*get_controls)(gpointer data);
	gpointer get_controls_data;
	const GList* (*get_buttons)(const char *control, gpointer data);
	gpointer get_buttons_data;
} Private;

#define GET_PRIVATE(obj) (Private*) g_object_get_data (G_OBJECT (obj), "preferences:priv")

static void
priv_destroy (gpointer data)
{
	Private *priv = (Private*)data;
	
	g_object_unref (G_OBJECT (priv->xml));
}



static void
load_clicked (GtkButton *button, Private *priv)
{
	const GList *controls;
	const char *control_name;

	controls = priv->get_controls (priv->get_controls_data);
	control_name = controls->data;

	while (controls) {
		const GList *buttons;
		const char *button_name;

		buttons = priv->get_buttons (control_name,
					     priv->get_buttons_data);

		while (buttons) {
			button_name = buttons->data;

			if (! gir_mappings_get_action (priv->mappings,
						       control_name,
						       button_name)) {
				gir_mappings_set_action_default
					(priv->mappings,
					 control_name,
					 button_name);
			}
			buttons = buttons->next;
		}
		
		controls = controls->next;
	}

	gir_mappings_sync (priv->mappings);
}

static void
treeview_popup (Private *priv, int x, int y, guint32 time)
{
	GladeXML *xml = priv->xml;
	GtkTreeView *tree_view;
	GtkMenu *menu;
	GtkWidget *edit_item, *delete_item;
	GtkTreePath *path;

	tree_view = GTK_TREE_VIEW (glade_xml_get_widget (xml, "treeview"));
	menu = GTK_MENU (glade_xml_get_widget (xml, "popup_menu"));
	edit_item = glade_xml_get_widget (xml, "popup_edit");
	delete_item = glade_xml_get_widget (xml, "popup_delete");

	gtk_widget_grab_focus (GTK_WIDGET (tree_view));
	if (gtk_tree_view_get_path_at_pos (tree_view, x, y, &path,
					   NULL, NULL, NULL)) {
		/* Select the row */
		if (gtk_tree_path_get_depth (path) == 2) {
			gtk_tree_selection_select_path (gtk_tree_view_get_selection (tree_view), path);
			gtk_widget_set_sensitive (edit_item, TRUE);
			gtk_widget_set_sensitive (delete_item, TRUE);
		} else {
			gtk_widget_set_sensitive (edit_item, FALSE);
			gtk_widget_set_sensitive (delete_item, FALSE);
		}
		gtk_tree_path_free (path);
	} else {
		gtk_widget_set_sensitive (edit_item, FALSE);
		gtk_widget_set_sensitive (delete_item, FALSE);
	}

	gtk_menu_popup (menu, NULL, NULL, NULL, NULL,
			3, time);

}

static gboolean
treeview_button_press_event (GtkTreeView *tree_view, GdkEventButton *event,
			     Private *priv)
{
	if (event->button == 3) {
		/* popup */
		treeview_popup (priv, event->x, event->y, event->time);
		return TRUE;
	}
	return FALSE;
}

static void
popup_new_activate (GtkMenuItem *item, Private *priv)
{
	g_warning ("not implemented");
}

static void
popup_edit_activate (GtkMenuItem *item, Private *priv)
{
	GtkWidget *dialog = glade_xml_get_widget (priv->xml, "edit_dialog");
	GtkTreeView *tree_view = priv->tree_view;
	GtkTreeModel *tree_model = priv->tree_model;
	GtkTreeSelection *selection = gtk_tree_view_get_selection (tree_view);
	GtkTreeIter iter;

	gtk_tree_selection_get_selected (selection, &tree_model, &iter);
	
	

	gtk_widget_show (dialog);
}

static void
popup_delete_activate (GtkMenuItem *item, Private *priv)
{
	g_warning ("not implemented");
}



static void
add_columns (GtkTreeView *tree_view)
{
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	int coloff;
	
	renderer = gtk_cell_renderer_text_new ();
	coloff = gtk_tree_view_insert_column_with_attributes
		(tree_view, -1, _("Signal"), renderer, "text",
		 SIGNAL_COLUMN, NULL);
	column = gtk_tree_view_get_column (tree_view, coloff);

	renderer = gtk_cell_renderer_text_new ();
	coloff = gtk_tree_view_insert_column_with_attributes
		(tree_view, -1, _("Action"), renderer, "text",
		 ACTION_COLUMN, NULL);
	column = gtk_tree_view_get_column (tree_view, coloff);
}

GtkWidget *
gir_preferences_new (GirMappings *mappings)
{
	GladeXML *xml;
	GtkWidget *dialog;
	GtkTreeView *tree_view;
	GtkTreeModel *tree_model;
	Private *priv;

	xml = glade_xml_new (GLADEDIR "/gir-preferences.glade", NULL, NULL);
	dialog = glade_xml_get_widget (xml, "preferences_dialog");

	tree_view = GTK_TREE_VIEW (glade_xml_get_widget (xml, "treeview"));
	add_columns (tree_view);

	priv = g_new0 (Private, 1);
	g_object_set_data_full (G_OBJECT (dialog), "preferences:priv", priv,
				priv_destroy);

	priv->xml = xml;
	priv->test_check_button = GTK_CHECK_BUTTON
		(glade_xml_get_widget (xml, "test_check_button"));
	priv->test_control_label = GTK_LABEL
		(glade_xml_get_widget (xml, "control_label"));
	priv->test_button_label = GTK_LABEL
		(glade_xml_get_widget (xml, "button_label"));
	priv->test_repeat_label = GTK_LABEL
		(glade_xml_get_widget (xml, "repeat_label"));

	priv->mappings = mappings;

	tree_model = gir_mappings_get_tree_model (mappings);
	gtk_tree_view_set_model (tree_view, tree_model);
	priv->tree_model = tree_model;

	glade_xml_signal_connect_data (xml, "load_clicked",
				       G_CALLBACK (load_clicked), priv);
	glade_xml_signal_connect_data (xml, "treeview_button_press_event",
				       G_CALLBACK (treeview_button_press_event), priv);
	glade_xml_signal_connect_data (xml, "popup_new_activate",
				       G_CALLBACK (popup_new_activate), priv);
	glade_xml_signal_connect_data (xml, "popup_edit_activate",
				       G_CALLBACK (popup_edit_activate), priv);
	glade_xml_signal_connect_data (xml, "popup_delete_activate",
				       G_CALLBACK (popup_delete_activate), priv);

	gtk_widget_show_all (dialog);

	return dialog;
}

gboolean
gir_preferences_is_test_mode (GtkWidget *pref)
{
	Private *priv = GET_PRIVATE (pref);
	return gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (priv->test_check_button));
}

void
gir_preferences_set_test_result (GtkWidget *pref,
				 const char *control, const char *button,
				 int repeat)
{
	Private *priv = GET_PRIVATE (pref);
	char *repeat_str;
	
	gtk_label_set_text (priv->test_control_label, control);
	gtk_label_set_text (priv->test_button_label, button);
	repeat_str = g_strdup_printf ("%d", repeat);
	gtk_label_set_text (priv->test_repeat_label, repeat_str);
	g_free (repeat_str);
}

void
gir_preferences_set_controls_callback (GtkWidget *pref,
				       const GList* (*callback)(gpointer data),
				       gpointer data)
{
	Private *priv = GET_PRIVATE (pref);

	priv->get_controls = callback;
	priv->get_controls_data = data;
}

void
gir_preferences_set_buttons_callback (GtkWidget *pref,
				      const GList* (*callback)(const char *c,
							       gpointer data),
				      gpointer data)
{
	Private *priv = GET_PRIVATE (pref);

	priv->get_buttons = callback;
	priv->get_buttons_data = data;
}

