/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
#include <config.h>

#include <gnome.h>
#include <gconf/gconf-client.h>

#include "gir-lirc.h"
#include "gir-mappings.h"
#include "gir-tray-icon.h"
#include "gir-preferences.h"


typedef struct {
	GirLirc *lirc;
	GirMappings *mappings;
	GirTrayIcon *tray_icon;

	GtkWidget *preferences;
	gboolean test_mode;
} MainContext;



#define GCONF_KEY_DIR "/apps/gir"
#define GCONF_KEY_CONNECT GCONF_KEY_DIR "/connect"



static void
gconf_notify_connect (GConfClient *client, guint cnxn_id,
		      GConfEntry *entry, gpointer data)
{
	MainContext *ctx = (MainContext *)data;
	gboolean active;

	active = gconf_value_get_bool (gconf_entry_get_value (entry));
	if (active)
		gir_lirc_connect (ctx->lirc, NULL); /* TODO: error handling */
	else
		gir_lirc_disconnect (ctx->lirc);
	gir_tray_icon_set_connected (ctx->tray_icon, active);
}




static void
preferences_closed (MainContext *ctx)
{
	gtk_widget_destroy (ctx->preferences);
	ctx->preferences = NULL;
}

static void
preferences_response (MainContext *ctx, int response_id)
{
	if (response_id == GTK_RESPONSE_CLOSE) {
		gtk_widget_destroy (ctx->preferences);
		ctx->preferences = NULL;
	}
}

static const GList *
preferences_get_controls (gpointer data)
{
	MainContext *ctx = (MainContext *)data;
	if (gir_lirc_is_connected (ctx->lirc))
		return gir_lirc_get_control_list (ctx->lirc);
	else
		return NULL;
}

static const GList *
preferences_get_buttons (const char *control, gpointer data)
{
	MainContext *ctx = (MainContext *)data;
	if (gir_lirc_is_connected (ctx->lirc))
		return gir_lirc_get_button_list (ctx->lirc, control);
	else
		return NULL;
}



static void
menu_connect (GirTrayIcon *tray_icon, gboolean active, gpointer data)
{
	GConfValue *value;

	value = gconf_value_new (GCONF_VALUE_BOOL);
	gconf_value_set_bool (value, active);
	gconf_client_set (gconf_client_get_default (),
			  GCONF_KEY_CONNECT, value, NULL);
	gconf_value_free (value);
}

static void
menu_about (GirTrayIcon *tray_icon, gpointer data)
{
	GtkWidget *about;
	const char *authors[] = {
		"\xeb\xa5\x98\xec\xb0\xbd\xec\x9a\xb0 <cwryu@debian.org>",
		NULL
	};
	const char *documenters[] = {
		NULL
	};
	char *translator_credits = _("translator-credits");

	translator_credits =
		strcmp (translator_credits, "translator-credits") ?
		translator_credits : NULL;

	about = gnome_about_new ("GIR",
				 PACKAGE_VERSION,
				 "Copyright \xc2\xa9 2004 Changwoo Ryu",
				 _("IR tool"),
				 authors,
				 documenters,
				 translator_credits,
				 NULL);

	gtk_widget_show (about);
}

static void
menu_preferences (GirTrayIcon *tray_icon, gpointer data)
{
	MainContext *ctx = (MainContext *)data;

	if (ctx->preferences) {
		gtk_window_present (GTK_WINDOW (ctx->preferences));
		return;
	}

	ctx->preferences = gir_preferences_new (ctx->mappings);
	g_signal_connect_swapped (G_OBJECT (ctx->preferences), "delete_event",
				  G_CALLBACK (preferences_closed), ctx);
	g_signal_connect_swapped (G_OBJECT (ctx->preferences), "response",
				  G_CALLBACK (preferences_response), ctx);

	gir_preferences_set_controls_callback (ctx->preferences,
					       preferences_get_controls,
					       ctx);
	gir_preferences_set_buttons_callback (ctx->preferences,
					      preferences_get_buttons,
					      ctx);
}

static void
menu_quit (GirTrayIcon *tray_icon, gpointer data)
{
	MainContext *ctx = (MainContext *)data;

	gir_tray_icon_free (ctx->tray_icon);
	gtk_main_quit ();
}



#define LIRC_RECONNECT_TIMEOUT 1000

static gboolean
lirc_reconnect_timeout (gpointer data)
{
	MainContext *ctx = (MainContext *)data;

	if (gir_lirc_is_connected (ctx->lirc))
		return FALSE;	/* already connected */

	if (gir_lirc_connect (ctx->lirc, NULL))
		return FALSE;

	return TRUE;		/* try later */
}

static void
lirc_recv_ir (MainContext *ctx,
	      int repeat_count, const char *button, const char *control)
{
	/* blink the icon */
	gir_tray_icon_set_active (ctx->tray_icon);

	if (ctx->preferences &&
	    gir_preferences_is_test_mode (ctx->preferences)) {
		/* test mode */
		gir_preferences_set_test_result (ctx->preferences, control,
						 button, repeat_count);
		return;
	}

	gir_mappings_do_action (ctx->mappings, control, button, repeat_count);
}

static void
lirc_disconnected (MainContext *ctx)
{
	g_timeout_add (LIRC_RECONNECT_TIMEOUT, lirc_reconnect_timeout, ctx);
}



int
main (int argc, char *argv[])
{
	MainContext ctx;
	GirLirc *lirc;
	GirTrayIcon *tray_icon;
	GConfValue *value;
	
	bindtextdomain (GETTEXT_PACKAGE, LOCALEDIR);
	bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
	textdomain (GETTEXT_PACKAGE);

	gnome_program_init (PACKAGE_NAME, PACKAGE_VERSION,
			    LIBGNOMEUI_MODULE, argc, argv,
			    NULL);

	bzero (&ctx, sizeof(ctx));

	ctx.lirc = lirc = gir_lirc_new ();
	ctx.mappings = gir_mappings_new ();

	ctx.tray_icon = tray_icon = gir_tray_icon_new ();

	g_signal_connect_swapped (G_OBJECT (lirc), "recv_ir",
				  G_CALLBACK (lirc_recv_ir), &ctx);
	g_signal_connect_swapped (G_OBJECT (lirc), "disconnected",
				  G_CALLBACK (lirc_disconnected), &ctx);

	gir_tray_icon_set_connect_callback (tray_icon, menu_connect, &ctx);
	gir_tray_icon_set_prefs_callback (tray_icon, menu_preferences, &ctx);
	gir_tray_icon_set_about_callback (tray_icon, menu_about, &ctx);
	gir_tray_icon_set_quit_callback (tray_icon, menu_quit, &ctx);

	gconf_client_add_dir (gconf_client_get_default (),
			      GCONF_KEY_DIR,
			      GCONF_CLIENT_PRELOAD_ONELEVEL,
			      NULL);

	value = gconf_client_get (gconf_client_get_default (),
				  GCONF_KEY_CONNECT,
				  NULL);
	if (gconf_value_get_bool (value)) {
		gir_lirc_connect (ctx.lirc, NULL);
		gir_tray_icon_set_connected (ctx.tray_icon, TRUE);
	}
	gconf_value_free (value);
	gconf_client_notify_add (gconf_client_get_default (),
				 GCONF_KEY_CONNECT,
				 gconf_notify_connect,
				 &ctx,
				 NULL, NULL);

	gtk_main ();
	return 0;
}

